package eu.clarin.cmdi.oai.harvester.impl;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.stream.Location;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import org.apache.http.HttpEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.codehaus.stax2.XMLStreamReader2;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import eu.clarin.cmdi.oai.harvester.HarvesterException;
import eu.clarin.cmdi.oai.harvester.HarvesterProtocolErrorException;
import eu.clarin.cmdi.oai.harvester.ProtocolError;


final class Response {
    private static final class CountingInputStream extends FilterInputStream {
        private long count = 0;


        private CountingInputStream(final InputStream stream) {
            super(stream);
        }


        @Override
        public int read() throws IOException {
            final int result = super.read();
            count += (result > 0) ? 1 : 0;
            return result;
        }


        @Override
        public int read(byte[] buffer, int offset, int length)
                throws IOException {
            final int result = super.read(buffer, offset, length);
            count += (result > 0) ? result : 0;
            return result;
        }


        @Override
        public int read(byte[] buffer) throws IOException {
            final int result = super.read(buffer);
            count += (result > 0) ? result : 0;
            return result;
        }


        @Override
        public long skip(long n) throws IOException {
            final long result = super.skip(n);
            count += (result > 0) ? n : 0;
            return result;
        }


        public long getByteCount() {
            return count;
        }

    } // class CountingInputStream
    private static final String OAI_NS = "http://www.openarchives.org/OAI/2.0/";
    private static final String ENCODING_GZIP = "gzip";
    private static final String ENCODING_DEFLATE = "defalte";
    private static final Logger logger =
            LoggerFactory.getLogger(Response.class);
    private final CloseableHttpResponse response;
    private final long now = System.currentTimeMillis();
    private final CountingInputStream stream;
    private final XMLStreamReader2 reader;


    Response(HarvestJobImpl job, CloseableHttpResponse response, AbstractHarvester harvester)
            throws IOException, XMLStreamException, HarvesterException {
        if (job == null) {
           throw new NullPointerException("job == null");
        }
        if (response == null) {
            throw new NullPointerException("response == null");
        }
        if (harvester == null) {
            throw new NullPointerException("harvester == null");
        }
        this.response = response;

        final HttpEntity entity = response.getEntity();
        if (entity.getContentType() != null) {
            logger.debug("Content-Type: {}", entity.getContentType().getValue());
            logger.debug("Content-Length: {}", entity.getContentLength());
        }
        InputStream in             = entity.getContent();
        org.apache.http.Header enc = entity.getContentEncoding();
        if (enc != null) {
            final String encoding = enc.getValue();
            logger.debug("Content-Encoding: {}", encoding);
            if (ENCODING_GZIP.equalsIgnoreCase(encoding)) {
                in = new GZIPInputStream(in);
            } else if (ENCODING_DEFLATE.equalsIgnoreCase(encoding)) {
                in = new InflaterInputStream(in, new Inflater(true));
            } else {
                throw new HarvesterException("Unsupported content encoding " +
                        "in HTTP response: " + encoding);
            }
        }
        this.stream = new CountingInputStream(job.wrap(in));
        this.reader = harvester.createReader(stream);
        readStart(OAI_NS, "OAI-PMH", true);
        readContent(OAI_NS, "responseDate", true);
        readStart(OAI_NS, "request", true, true);
        readEnd(OAI_NS, "request", true);

        List<ProtocolError> errors = null;
        while (readStart(OAI_NS, "error", false, true)) {
            if (errors == null) {
                errors = new ArrayList<ProtocolError>();
            }
            final String code = reader.getAttributeValue(null, "code");
            reader.next(); // skip start element after reading attrs
            final String msg = readString(true);
            readEnd(OAI_NS, "error");
            errors.add(new ProtocolError(code, msg));
        }
        if (errors != null) {
            throw new HarvesterProtocolErrorException(errors);
        }
    }


    public void close() throws XMLStreamException {
        readEnd(OAI_NS, "OAI-PMH");
    }


    public void release(HarvestJobImpl job) {
        job.finishRequest(stream.getByteCount(),
                (System.currentTimeMillis() - now));
        try {
            reader.close();
        } catch (XMLStreamException e) {
            /* IGNORE */
        }
        try {
            stream.close();
        } catch (IOException e) {
            /* IGNORE */
        }

        /* make sure to release allocated resources */
        try {
            response.close();
        } catch (IOException e) {
            /* IGNORE */
        }
    }


    XMLStreamReader getXMLStreamReader() {
        return reader;
    }


    Location getLocation() {
        return reader.getLocation();
    }


    boolean readStart(String namespaceURI, String localName, boolean required)
            throws XMLStreamException {
        return readStart(namespaceURI, localName, required, false);
    }


    boolean readStart(String namespaceURI, String localName, boolean required,
            boolean attributes) throws XMLStreamException {
        // System.err.println("readStart (" + localName + ", required = " +
        // required + ") @ " + toReadable(reader));
        if (!reader.isEndElement()) {
            while (reader.hasNext()) {
                // System.err.println("  LOOP: " + dumpState());
                if (reader.isWhiteSpace()) {
                    reader.next();
                    continue;
                }
                if (reader.isStartElement()) {
                    if (namespaceURI.equals(reader.getNamespaceURI()) &&
                            localName.equals(reader.getLocalName())) {
                        // System.err.print("--> found ");
                        if (!attributes) {
                            // System.err.print("and consumed ");
                            reader.next(); // skip to next event
                        }
                        // System.err.println("@ " + toReadable(reader));
                        return true;
                    }
                    break;
                }
                if (reader.isCharacters() || reader.isEndElement()) {
                    break;
                }
                reader.next();
            } // while
        }
        if (required) {
            // System.err.println("--> error, not found @ " +
            // toReadable(reader));
            String what;
            if (reader.isStartElement() || reader.isEndElement() ||
                    reader.isEmptyElement()) {
                what = "'" + reader.getName().toString() + "'";
            } else {
                what = "some character data";
            }
            throw new XMLStreamException("expected element '" +
                    new QName(namespaceURI, localName) + "', but found "+
                    what, reader.getLocation());
        }
        // System.err.println("--> not found @ " + toReadable(reader));
        return false;
    }


    void readEnd(String namespaceURI, String localName)
            throws XMLStreamException {
        readEnd(namespaceURI, localName, false);
    }


    void readEnd(String namespaceURI, String localName, boolean skipContent)
            throws XMLStreamException {
        // System.err.println("readEnd (" + localName + ") @ " + dumpState() +
        // ", skipContent = " + skipContent);
        int level = 1;
        while (reader.hasNext()) {
            // System.err.println("  LOOP " + dumpState() + " [" +
            // level + "]");
            if (reader.isWhiteSpace()) {
                reader.next();
                continue;
            }
            if (skipContent) {
                if (reader.isCharacters()) {
                    reader.next();
                    continue;
                }
                if (reader.isStartElement()) {
                    if (!(namespaceURI.equals(reader.getNamespaceURI()) &&
                            localName.equals(reader.getLocalName()))) {
                        level++;
                    }
                    reader.next();
                    continue;
                }
            }
            if (reader.isEndElement()) {
                level--;
                // System.err.println("   @END-TAG: " + dumpState() + " [" +
                // level + "]");
                if (level == 0) {
                    if (namespaceURI.equals(reader.getNamespaceURI()) &&
                            localName.equals(reader.getLocalName())) {
                        reader.next(); // consume tag
                        break;
                    } else {
                        String what;
                        if (reader.isStartElement() || reader.isEndElement() ||
                                reader.isEmptyElement()) {
                            what = "'" + reader.getName().toString() + "'";
                        } else {
                            what = "some character data";
                        }
                        throw new XMLStreamException("expected end tag for '" +
                                new QName(namespaceURI, localName) +
                                "', but found " + what,
                                reader.getLocation());
                    }
                }
            }
            reader.next();
        }
        // System.err.println("--> ok @ " + dumpState());
    }


    boolean peekStart(String namespaceURI, String localName)
            throws XMLStreamException {
        if (reader.isWhiteSpace()) {
            consumeWhitespace();
        }
        if (!reader.isStartElement()) {
            return false;
        }
        return namespaceURI.equals(reader.getNamespaceURI()) &&
                localName.equals(reader.getLocalName());
    }


    String readContent(String namespaceURI, String localName, boolean required)
            throws XMLStreamException {
        return readContent(namespaceURI, localName, required, true);
    }


    String readContent(String namespaceURI, String localName, boolean required,
            boolean contentRequired) throws XMLStreamException {
        String result = null;
        if (readStart(namespaceURI, localName, required)) {
            try {
                result = readString(contentRequired);
                if (!contentRequired && (result == null)) {
                    result = "";
                }
            } catch (XMLStreamException e) {
                StringBuilder sb = new StringBuilder();
                sb.append("element '");
                sb.append(new QName(namespaceURI, localName));
                sb.append("' may not be empty");
                throw new XMLStreamException(sb.toString(), e.getLocation());
            }
            readEnd(namespaceURI, localName);
        }
        return result;
    }


    int readContent(String namespaceURI, String localName, boolean required,
            int defaultValue) throws XMLStreamException {
        if (readStart(namespaceURI, localName, required)) {
            String s = readString(true);
            try {
                readEnd(namespaceURI, localName);
            } catch (XMLStreamException e) {
                StringBuilder sb = new StringBuilder();
                sb.append("element '");
                sb.append(new QName(namespaceURI, localName));
                sb.append(localName).append("' may not be empty");
                throw new XMLStreamException(sb.toString(), e.getLocation());
            }
            try {
                return Integer.parseInt(s);
            } catch (NumberFormatException e) {
                StringBuilder sb = new StringBuilder();
                sb.append("element '");
                sb.append(new QName(namespaceURI, localName));
                sb.append("' was expected to be of type xs:integer; ");
                sb.append("incompatible value was: ");
                sb.append(s);
                throw new XMLStreamException(sb.toString(),
                        reader.getLocation(), e);
            }
        }
        return defaultValue;
    }


    String readString(boolean required) throws XMLStreamException {
        // System.err.println("readString @ " + toReadable(reader));
        StringBuilder sb = new StringBuilder();
        while (reader.isCharacters()) {
            String s = reader.getText();
            if ((s != null) && !s.isEmpty()) {
                sb.append(s);
            }
            if (!reader.hasNext()) {
                break;
            }
            reader.next();
        } // while
        String s = null;
        if (sb.length() > 0) {
            s = sb.toString().trim();
        }
        if (required && ((s == null) || s.isEmpty())) {
            throw new XMLStreamException("expected character content "
                    + "at position ", reader.getLocation());
        }
        // System.err.println("--> ok @ " + toReadable(reader));
        return s;
    }


    String readAttributeValue(String namespaceURI, String localName)
            throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        String attr = reader.getAttributeValue(namespaceURI, localName);
        if (attr != null) {
            attr = attr.trim().intern();
        }
        return attr;
    }


    String readNamespaceURI() throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        return reader.getNamespaceURI();
    }


    String peekElementLocalName() throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        return reader.getLocalName();
    }


    void consumeStart() throws XMLStreamException {
        if (!reader.isStartElement()) {
            throw new XMLStreamException("not at a start elment event",
                    reader.getLocation());
        }
        reader.next();
    }


    void consumeWhitespace() throws XMLStreamException {
        outer:
        while (reader.hasNext()) {
            /*
             * some hackish method to also skip stray UTF-8 BOMs; only works
             * where whitespace is expected
             */
            if (reader.isCharacters()) {
                final char buffer[] = reader.getTextCharacters();
                final int end = reader.getTextStart() + reader.getTextLength();
                for (int i = reader.getTextStart(); i < end; i++) {
                    if (!(Character.isWhitespace(buffer[i]) ||
                            (buffer[i] == 0x0FEFF))) {
                        break outer;
                    }
                }
            } else if (!reader.isWhiteSpace()) {
                break outer;
            }
            reader.next();
        } // while
    }


    void copyTo(XMLStreamWriter writer) throws XMLStreamException {
        final int depth = reader.getDepth();
        do {
            copyEvent(reader, writer);
            reader.next();
        } while (reader.getDepth() >= depth);
    }


    private static void copyEvent(XMLStreamReader from, XMLStreamWriter to)
            throws XMLStreamException {
        switch (from.getEventType()) {
        case XMLStreamConstants.START_DOCUMENT:
            {
                String version = from.getVersion();
                if (version == null || version.length() == 0) {
                    to.writeStartDocument();
                } else {
                    to.writeStartDocument(from.getCharacterEncodingScheme(),
                            from.getVersion());
                }
                to.writeCharacters("\n");
            }
            return;

        case XMLStreamConstants.END_DOCUMENT:
            to.writeCharacters("\n");
            to.writeEndDocument();
            return;

        case XMLStreamConstants.START_ELEMENT:
            copyStartElement(from, to);
            return;

        case XMLStreamConstants.END_ELEMENT:
            to.writeEndElement();
            return;

        case XMLStreamConstants.SPACE:
            to.writeCharacters(from.getTextCharacters(), from.getTextStart(),
                    from.getTextLength());
            return;

        case XMLStreamConstants.CDATA:
            to.writeCData(from.getText());
            return;

        case XMLStreamConstants.CHARACTERS:
            to.writeCharacters(from.getTextCharacters(), from.getTextStart(),
                    from.getTextLength());
            return;

        case XMLStreamConstants.COMMENT:
            to.writeComment(from.getText());
            return;

        case XMLStreamConstants.PROCESSING_INSTRUCTION:
            to.writeProcessingInstruction(from.getPITarget(), from.getPIData());
            return;

        case XMLStreamConstants.DTD:
        case XMLStreamConstants.ENTITY_REFERENCE:
        case XMLStreamConstants.ATTRIBUTE:
        case XMLStreamConstants.NAMESPACE:
        case XMLStreamConstants.ENTITY_DECLARATION:
        case XMLStreamConstants.NOTATION_DECLARATION:
            /* FALL_TROUGH */
        }
        throw new XMLStreamException("unsupported event type: " +
                from.getEventType());
    }


    private static void copyStartElement(XMLStreamReader from,
            XMLStreamWriter to) throws XMLStreamException {
        final int nsCount = from.getNamespaceCount();
        if (nsCount > 0) { // yup, got some...
            for (int i = 0; i < nsCount; ++i) {
                String pfx = from.getNamespacePrefix(i);
                String uri = from.getNamespaceURI(i);
                if ((pfx == null) || pfx.isEmpty()) { // default NS
                    to.setDefaultNamespace(uri);
                } else {
                    to.setPrefix(pfx, uri);
                }
            }
        }

        final String prefix = from.getPrefix();
        final NamespaceContext from_ctx = from.getNamespaceContext();
        final NamespaceContext to_ctx = to.getNamespaceContext();
        boolean repair_prefix_namespace = false;
        if ((prefix != null) && (to_ctx.getNamespaceURI(prefix) == null)) {
            repair_prefix_namespace = true;
            to.setPrefix(prefix, from_ctx.getNamespaceURI(prefix));
        }

        to.writeStartElement(prefix, from.getLocalName(),
                from.getNamespaceURI());

        if (nsCount > 0) {
            // write namespace declarations
            for (int i = 0; i < nsCount; ++i) {
                String pfx = from.getNamespacePrefix(i);
                String uri = from.getNamespaceURI(i);

                if ((pfx == null) || pfx.isEmpty()) { // default NS
                    to.writeDefaultNamespace(uri);
                } else {
                    to.writeNamespace(pfx, uri);
                }
            }
        }
        if (repair_prefix_namespace) {
            to.writeNamespace(prefix, from_ctx.getNamespaceURI(prefix));
        }

        int attrCount = from.getAttributeCount();
        if (attrCount > 0) {
            for (int i = 0; i < attrCount; ++i) {
                to.writeAttribute(from.getAttributePrefix(i),
                        from.getAttributeNamespace(i),
                        from.getAttributeLocalName(i),
                        from.getAttributeValue(i));
            }
        }
    }


    String dumpState() {
        StringBuilder sb = new StringBuilder();
        switch (reader.getEventType()) {
        case XMLStreamConstants.START_DOCUMENT:
            return "START_DOC[";
        case XMLStreamConstants.END_DOCUMENT:
            return "END_DOC[";
        case XMLStreamConstants.START_ELEMENT:
            sb.append("START[");
            sb.append(reader.getNamespaceURI());
            sb.append(",");
            sb.append(reader.getLocalName());
            break;
        case XMLStreamConstants.END_ELEMENT:
            sb.append("END[");
            sb.append(reader.getNamespaceURI());
            sb.append(",");
            sb.append(reader.getLocalName());
            break;
        case XMLStreamConstants.CHARACTERS:
            sb.append("CHARACTERS[\"");
            sb.append(reader.getText()
                    .replace("\n", "\\n")
                    .replace("\r", "\\r")
                    .replace("\t", "\\t"));
            sb.append("\", isWhitespace = ");
            sb.append(reader.isWhiteSpace());
            break;
        case XMLStreamConstants.CDATA:
            sb.append("CDATA[\"");
            sb.append(reader.getText().replace("\n", "\\n")
                    .replace("\r", "\\r").replace("\t", "\\t"));
            sb.append("\", isWhitespace = ");
            sb.append(reader.isWhiteSpace());
            break;
        default:
            sb.append(Integer.toString(reader.getEventType()));
        }
        Location location = reader.getLocation();
        sb.append(", row = ").append(location.getLineNumber());
        sb.append(", col = ").append(location.getColumnNumber());
        sb.append(", offset = ").append(location.getCharacterOffset());
        sb.append("]");
        return sb.toString();
    }

} // class XmlStreamReaderWrapper
